// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32-COS/HPDI32-COS-1.x_GSC_DN/driver/device.c $
// $Rev: 54374 $
// $Date: 2024-04-30 16:41:22 -0500 (Tue, 30 Apr 2024) $

// HPDI32-COS: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type

	{ "HPDI32-COS",	0x10B5, 0x9080, 0x10B5, 0x2400,	GSC_DEV_TYPE_HPDI32COS	},
	{ NULL }

	// NOTE: 9080/2400 id values are also used by generic HPDI32 model boards.
	// However, COS boards have 0x10 in FRR.D16-D23, which is checked below.
};



//*****************************************************************************
static int _check_support(dev_data_t* dev)
{
	u32		frr;
	VADDR_T	frr_va;
	int		ret		= 0;

	// HPDI32-COS models have 0x10 in bits D16-D23
	frr_va	= GSC_VADDR(dev, 0x00);
	frr		= os_reg_mem_rx_u32(dev, frr_va);

	if ((frr & 0x00FF0000) != 0x00100000)
	{
		ret	= -ENODEV;	// This is NOT one of our devices.
#if DEV_PCI_ID_SHOW
		printf("%s: This device is not supported by this driver.\n", DEV_NAME);
		printf("%s: This is not an HPDI32-COS model device.\n", DEV_NAME);
		printf("%s: FRR 0x%08lX\n", DEV_NAME, (long) frr);
#endif
	}

	return(ret);
}



//*****************************************************************************
static void _fifo_compute(dev_data_t* dev)
{
	dev->cache.fifo_size_rx	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_rfsr_32);
}



//*****************************************************************************
static void _form_factor_compute(dev_data_t* dev)
{
	if (dev->cache.model == HPDI32COS_MODEL_HPDI32A)
	{
		dev->cache.form_factor	= HPDI32COS_FORM_FACTOR_UNKNOWN;
	}
	else
	{
		dev->cache.form_factor	= HPDI32COS_FORM_FACTOR_PMC;
	}
}



//*****************************************************************************
static void _model_compute(dev_data_t* dev)
{
	if ((dev->cache.gsc_frr_32 & 0xFFF) <= 0x201)
	{
		dev->model			= "HPDI32A-COS";
		dev->cache.model	= HPDI32COS_MODEL_HPDI32A;
	}
	else
	{
		dev->model			= "HPDI32B-COS";
		dev->cache.model	= HPDI32COS_MODEL_HPDI32B;
	}
}



//*****************************************************************************
static void _proc_id_str_compute(dev_data_t* dev)
{
	sprintf(dev->proc_id_str, "0x%lX", (long) dev->cache.jumper_val);
}



//*****************************************************************************
static void _user_jumpers_compute(dev_data_t* dev)
{
	u32	bsr;
	u32	val;

	bsr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bsr_32);
	val	= (bsr >> 16) & 0x3;

	dev->cache.jumper_on	= 1;
	dev->cache.jumper_qty	= 2;
	dev->cache.jumper_val	= val;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(HPDI32COS_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Verify that this is a supported device.
		ret	= _check_support(dev);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_frr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_FRR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_BCR));
		dev->vaddr.gsc_bsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_BSR));
		dev->vaddr.gsc_ecr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_ECR));
		dev->vaddr.gsc_rar_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RAR));
		dev->vaddr.gsc_scdr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_SCDR));
		dev->vaddr.gsc_rdfr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RDFR));
		dev->vaddr.gsc_tdor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_TDOR));
		dev->vaddr.gsc_rdir_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RDIR));
		dev->vaddr.gsc_latr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_LATR));
		dev->vaddr.gsc_dmr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_DMR));
		dev->vaddr.gsc_ecivr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_ECIVR));
		dev->vaddr.gsc_rwcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RWCR));
		dev->vaddr.gsc_icr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_ICR));
		dev->vaddr.gsc_isr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_ISR));
		dev->vaddr.gsc_rfsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RFSR));
		dev->vaddr.gsc_rfwcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_RFWCR));
		dev->vaddr.gsc_ielr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_IELR));
		dev->vaddr.gsc_ihlr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(HPDI32COS_GSC_IHLR));

		// Data cache initialization.

		dev->cache.gsc_frr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_frr_32);

		dev->cache.clock_div_max	= 0xFFFF;
		dev->cache.fref_default		= 20000000;
		dev->cache.fref_max			= 32000000;
		dev->cache.init_ms			= 1;

		_model_compute(dev);		// 1st
		_form_factor_compute(dev);	// 2nd
		_fifo_compute(dev);			// 3rd
		_user_jumpers_compute(dev);
		_proc_id_str_compute(dev);	// last

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_DMDMA_READ;
		ret	= gsc_dma_create(dev, dma, dma);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


