// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32-COS/HPDI32-COS-1.x_GSC_DN/driver/read.c $
// $Rev: 54346 $
// $Date: 2024-04-30 14:42:58 -0500 (Tue, 30 Apr 2024) $

// HPDI32-COS: Device Driver: source file

#include "main.h"



//*****************************************************************************
static long _rx_available(dev_data_t* dev, size_t bytes)
{
	u32	almost;
	u32	bsr;
	u32	fwr;
	u32	rar;

	for (;;)	/* A convenience loop. */
	{
		// The Rx FIFO Words Register is available.

		if (dev->cache.reg_fwr_update)
		{
			// In some versions of firmware the
			// RFWR may return zero when full.
			fwr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_rfwcr_32);

			if (fwr)
			{
				// The Rx FIFO has data.
				// Available = RFWR
				bytes	= fwr;
				break;
			}

			bsr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bsr_32);

			if ((bsr & D15) == 0)
			{
				// The Rx FIFO is Full.
				// Available = Rx FIFO Size
				bytes	= dev->cache.fifo_size_rx;
				break;
			}

			// The Rx FIFO is Empty.
			// Available = 0
			bytes	= 0;
			break;
		}

		bsr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bsr_32);

		// The Rx FIFO Size Register is available.

		if ((bsr & 0x8000) == 0)
		{
			// The Rx FIFO is full.
			// Available = Rx FIFO Size
			bytes	= dev->cache.fifo_size_rx;
			break;
		}

		if ((bsr & 0x4000) == 0)
		{
			// The Rx FIFO is Almost Full.
			// Available >= Rx FIFO Size - Rx Almost Full
			rar		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_rar_32);
			almost	= GSC_FIELD_DECODE(rar,31,16);
			bytes	= dev->cache.fifo_size_rx - almost;
			break;
		}

		if (bsr & 0x2000)
		{
			// The Rx FIFO is not Almost Empty.
			// Available > Rx Almost Empty
			rar		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_rar_32);
			almost	= GSC_FIELD_DECODE(rar,15,0);
			bytes	= almost + 1;
			break;
		}

		if (bsr & 0x1000)
		{
			// The Rx FIFO is not Empty.
			// Available >= 1
			bytes	= 1;
			break;
		}

		// The Rx FIFO is Empty.
		// Available = 0
		bytes	= 0;
		break;
	}

	bytes	*= dev->io.rx.bytes_per_sample;
	return((long) bytes);
}



//*****************************************************************************
static void _dev_io_sw_init(dev_data_t* dev, dev_io_t* io)
{
	io->bmdma_threshold	= HPDI32COS_RX_IO_BMDMA_THRESH_DEF;
	io->io_mode			= HPDI32COS_IO_MODE_DEFAULT;
	io->over_check		= HPDI32COS_IO_ERROR_DEFAULT;
	io->pio_threshold	= HPDI32COS_IO_PIO_THRESH_DEFAULT;
	io->timeout_s		= HPDI32COS_IO_TIMEOUT_DEFAULT;
	io->under_check		= HPDI32COS_IO_ERROR_DEFAULT;
}



//*****************************************************************************
static void _dev_io_close(dev_data_t* dev, dev_io_t* io)
{
	io->dma_channel	= NULL;
	os_mem_dma_close(&io->mem);
}



//*****************************************************************************
static void _dev_io_open(dev_data_t* dev, dev_io_t* io)
{
	_dev_io_sw_init(dev, io);
	os_mem_dma_open(&io->mem);
}



//*****************************************************************************
static int _dev_io_startup(dev_data_t* dev, dev_io_t* io)
{
	u32		bsr;
	long	ret		= 0;

	// Check for Rx FIFO overflows and underflows.

	if (io->over_check || io->under_check)
	{
		bsr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bsr_32);

		if ((io->under_check) && (bsr & D20))
			ret	= -EIO;

		if ((io->over_check) && (bsr & D19))
			ret	= -EIO;
	}

	return(ret);
}



//*****************************************************************************
static long _dev_pio_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	bytes	= _rx_available(dev, bytes);
	return((long) bytes);
}



//*****************************************************************************
static long _dev_pio_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	qty	= gsc_read_pio_work(dev, io, mem, bytes, st_end);
	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	s32	available;

	available	= _rx_available(dev, bytes);

	if (available >= (s32) bytes)
		bytes	= available;
	else if (available < io->bmdma_threshold)
		bytes	= 0;
	else
		bytes	= available;

	return((long) bytes);
}



//*****************************************************************************
static long _dev_bmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	long			samples	= (long) bytes / io->bytes_per_sample;
	gsc_dma_setup_t	setup;

	if (samples <= io->pio_threshold)
	{
		qty	= gsc_read_pio_work(dev, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= dev;
		setup.dev		= dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_READ;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE
						| GSC_DMA_MODE_SIZE_32_BITS;

		setup.dpr		= GSC_DMA_DPR_BOARD_TO_HOST
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(dev_data_t* dev, dev_io_t* io, size_t bytes)
{
	// Procede as if the input stream is continuous. If the input stream is not
	// continuous, then the application must alter its means of reading data.
	// The application may read the leading portion via DMDMA, but the trailing
	// Almost Empty sized portion must be read by DMA or PIO.
	return((long) bytes);
}



//*****************************************************************************
static long _dev_dmdma_xfer(
	dev_data_t*		dev,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty;
	long			samples	= (long) bytes / io->bytes_per_sample;
	gsc_dma_setup_t	setup;

	if (samples <= io->pio_threshold)
	{
		qty	= gsc_read_pio_work(dev, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= dev;
		setup.dev		= dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_DMDMA_READ;

		setup.mode		= GSC_DMA_MODE_DM_DMA
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE
						| GSC_DMA_MODE_SIZE_32_BITS;

		setup.dpr		= GSC_DMA_DPR_BOARD_TO_HOST
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	dev_read_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Read Input streaming I/O structure.
*
*	Arguments:
*
*		dev		The data for the device of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_read_create(dev_data_t* dev, dev_io_t* io)
{
	int	ret;

	io->bytes_per_sample	= 4;
	io->data_size_bits		= 32;
	io->io_reg_offset		= GSC_REG_OFFSET(HPDI32COS_GSC_RDFR);
	io->io_reg_vaddr		= dev->vaddr.gsc_rdfr_32;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available;
	io->dev_bmdma_available	= _dev_bmdma_available;
	io->dev_dmdma_available	= _dev_dmdma_available;
	io->dev_pio_xfer		= _dev_pio_xfer;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer;

	io->wait.abort			= HPDI32COS_WAIT_IO_RX_ABORT;
	io->wait.done			= HPDI32COS_WAIT_IO_RX_DONE;
	io->wait.error			= HPDI32COS_WAIT_IO_RX_ERROR;
	io->wait.timeout		= HPDI32COS_WAIT_IO_RX_TIMEOUT;

	ret	= gsc_io_create(dev, io, _8M);
	return(ret);
}


