// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32-COS/HPDI32-COS-1.x_GSC_DN/samples/irq/perform.c $
// $Rev: 54359 $
// $Date: 2024-04-30 14:55:26 -0500 (Tue, 30 Apr 2024) $

// HPDI32-COS: Sample Application: source file

#include "main.h"




// data types *****************************************************************

typedef struct
{
	const char*	name;
	s32			main;
	s32			gsc;
	s32			alt;
	s32			io;
	int			(*func)(int fd, int supported);
} irq_test_t;



// variables ******************************************************************

static	int			_fd;
static	os_sem_t	_sem_enter;
static	os_sem_t	_sem_exit;
static	int			_thread_done;
static	char		_thread_err[256];
static	char		_thread_info[256];
static	gsc_wait_t	_thread_wait;

static const irq_test_t	_tests[]	=
{
	// Main interrupts
	// name							main										gsc	alt	io	func
	{ "PCI",						GSC_WAIT_MAIN_PCI,							0,	0,	0,	main_pci_irq_test			},
	{ "DMA0/DMA1",					GSC_WAIT_MAIN_DMA0 | GSC_WAIT_MAIN_DMA1,	0,	0,	0,	main_dma_irq_test			},
	{ "GSC",						GSC_WAIT_MAIN_GSC,							0,	0,	0,	main_gsc_irq_test			},

	// GSC Interrupts
	// name							main	gsc										alt	io	func

	{ "CC1 Rise",					0,		HPDI32COS_WAIT_GSC_CC1_RISE,				0,	0,	cc1_rise_irq_test			},
	{ "CC1 Fall",					0,		HPDI32COS_WAIT_GSC_CC1_FALL,				0,	0,	cc1_fall_irq_test			},
	{ "CC2",						0,		HPDI32COS_WAIT_GSC_CC2,						0,	0,	cc2_irq_test				},
	{ "CC3",						0,		HPDI32COS_WAIT_GSC_CC3,						0,	0,	cc3_irq_test				},
	{ "CC4",						0,		HPDI32COS_WAIT_GSC_CC4,						0,	0,	cc4_irq_test				},
	{ "CC5",						0,		HPDI32COS_WAIT_GSC_CC5,						0,	0,	cc5_irq_test				},
	{ "CC6",						0,		HPDI32COS_WAIT_GSC_CC6,						0,	0,	cc6_irq_test				},

	{ "Rx FIFO Underflow",			0,		HPDI32COS_IRQ_RX_FIFO_UNDER,				0,	0,	rx_fifo_under_irq_test		},
	{ "Rx FIFO Overflow",			0,		HPDI32COS_WAIT_GSC_RX_FIFO_OVER,			0,	0,	rx_fifo_over_irq_test		},
	{ "Rx Stopped",					0,		HPDI32COS_WAIT_GSC_RX_STOPPED,				0,	0,	rx_stopped_irq_test			},
	{ "LA Triggered",				0,		HPDI32COS_WAIT_GSC_LA_TRIGGERED,			0,	0,	la_triggered_irq_test		},
	{ "Rx Running",					0,		HPDI32COS_WAIT_GSC_RX_RUNNING,				0,	0,	rx_running_irq_test			},
	{ "Event Count Zero",			0,		HPDI32COS_WAIT_GSC_EVENT_COUNT_0,			0,	0,	event_count_zero_irq_test	},
	{ "COS Detected",				0,		HPDI32COS_WAIT_GSC_COS_DETECTED,			0,	0,	cos_detected_irq_test		},

	{ "Rx FIFO Empty",				0,		HPDI32COS_WAIT_GSC_RX_FIFO_EMPTY,			0,	0,	rx_fifo_empty_irq_test		},
	{ "Rx FIFO AE",					0,		HPDI32COS_WAIT_GSC_RX_FIFO_AE,				0,	0,	rx_fifo_ae_irq_test			},
	{ "Rx FIFO AF",					0,		HPDI32COS_WAIT_GSC_RX_FIFO_AF,				0,	0,	rx_fifo_af_irq_test			},
	{ "Rx FIFO Full",				0,		HPDI32COS_WAIT_GSC_RX_FIFO_FULL,			0,	0,	rx_fifo_full_irq_test		},

	{ NULL }
};



//*****************************************************************************
static int _wait_thread(void* arg)
{
	int			errs;
	irq_test_t*	ptr		= (void*) arg;

	os_sem_unlock(&_sem_enter);

	for (;;)
	{
		memset(&_thread_wait, 0, sizeof(_thread_wait));
		_thread_wait.main		= ptr->main;
		_thread_wait.gsc		= ptr->gsc;
		_thread_wait.alt		= ptr->alt;
		_thread_wait.io			= ptr->io;
		_thread_wait.timeout_ms	= 15000;

		if (_thread_wait.main & GSC_WAIT_MAIN_PCI)
			_thread_wait.main	|= GSC_WAIT_MAIN_ALL;

		errs	= hpdi32cos_wait_event(_fd, -1, 0, &_thread_wait);

		if (errs)
		{
			sprintf(_thread_err, "%d. Wait request failed.", __LINE__);
			break;
		}

		if (_thread_wait.main & GSC_WAIT_MAIN_PCI)
		{
			// Since our PCI interrupts are generatred by DMA and GSC
			// operations we ignore all other PCI type events.

			if (_thread_wait.main & GSC_WAIT_MAIN_DMA0)
				break;

			if (_thread_wait.main & GSC_WAIT_MAIN_DMA1)
				break;

			if (_thread_wait.main & GSC_WAIT_MAIN_GSC)
				break;

			continue;
		}

		break;
	}

	sprintf(_thread_info,
			"M 0x%lX, G 0x%lX, A 0x%lX, I 0x%lX, T %ld\n",
			(long) _thread_wait.main,
			(long) _thread_wait.gsc,
			(long) _thread_wait.alt,
			(long) _thread_wait.io,
			(long) _thread_wait.timeout_ms);

	os_sem_unlock(&_sem_exit);

	_thread_done	= 1;
	return(errs);
}



//*****************************************************************************
static int _wait_for_waiting(int fd, const irq_test_t* ptr)
{
	int			errs	= 0;
	int			i;
	int			limit	= 1000;
	gsc_wait_t	wait;

	memset(&wait, 0, sizeof(wait));
	wait.main	= ptr->main;
	wait.gsc	= ptr->gsc;
	wait.alt	= ptr->alt;
	wait.io		= ptr->io;

	for (i = 0;; i++)
	{
		os_sleep_ms(10);

		if (_thread_done)
		{
			errs++;
			printf("FAIL <---  (%d. Thread resumed prematurely.)\n", __LINE__);

			gsc_label_level_inc();

			gsc_label("Thread Error");
			printf("%s\n", _thread_err);

			gsc_label("Thread Info");
			printf("%s\n", _thread_info);

			gsc_label_level_dec();
			break;
		}

		errs	= hpdi32cos_wait_status(fd, -1, 0, &wait);

		if (errs)
		{
			printf("FAIL <---  (%d. Wait status request failed.)\n", __LINE__);
			break;
		}
		else if (i > limit)
		{
			printf("FAIL <---  (%d. Excessive delay for thread to wait.)\n", __LINE__);
			errs++;
			break;
		}
		else if (wait.count)
		{
			break;
		}
	}

	return(errs);
}



//*****************************************************************************
static int _wait_for_done(void)
{
	int	errs	= 0;
	int	i;

	for (i = 0;; i++)
	{
		if (_thread_done)
			break;

		if (i > 1000)
		{
			printf("FAIL <---  (%d. Excessive delay for thread to complete.)\n", __LINE__);
			errs++;
			break;
		}
		else
		{
			os_sleep_ms(10);
		}
	}

	return(errs);
}



//*****************************************************************************
static int _wait_cancel(int fd)
{
	int			errs	= 0;
	gsc_wait_t	wait;

	memset(&wait, 0, sizeof(wait));
	wait.main	= 0xFFFFFFFF;
	wait.gsc	= 0xFFFFFFFF;
	wait.alt	= 0xFFFFFFFF;
	wait.io		= 0xFFFFFFFF;
	errs		+= hpdi32cos_wait_cancel(fd, -1, 0, &wait);

	return(errs);
}



//*****************************************************************************
static int _test_irq(int fd, const irq_test_t* ptr)
{
	int			errs		= 0;
	int			ret;
	os_thread_t	thread;

	for (;;)	// A cenvenience loop.
	{
		gsc_label(ptr->name);

		// Prepare for testing the interrupt.
		_thread_done	= 0;
		_thread_err[0]	= 0;
		_thread_info[0]	= 0;
		_fd				= fd;
		memset(&_thread_wait, 0, sizeof(_thread_wait));
		os_sem_create_qty(&_sem_enter, 1, 0);
		os_sem_create_qty(&_sem_exit, 1, 0);

		// Find out what level of support there is.
		ret	= ptr->func(fd, 1);

		if (ret > 0)
		{
			errs++;
			printf("FAIL <---  (%d. Request for support status failed.)\n", __LINE__);
			break;
		}

		if (ret == STATUS_NOT_SUPPORTED)
		{
			printf("SKIPPED  (%d. Interrupt not supported on this device.)\n", __LINE__);
			break;
		}

		if (ret == STATUS_NOT_IMPLEMENTED)
		{
			printf("SKIPPED  (%d. Interrupt test not implemented.)\n", __LINE__);
			break;
		}

		if (ret)
		{
			errs++;
			printf("SKIPPED  (%d. Invalid support status returned: %d)\n", __LINE__, ret);
			break;
		}

		// Create a thread to wait on the interrupt.
		errs	= os_thread_create(&thread, "wait thtread", _wait_thread, (void*) ptr);

		if (errs)
		{
			printf("FAIL <---  (%d. Thread creation failed.)\n", __LINE__);
			break;
		}

		// Wait for the thread to start running.
		ret	= os_sem_lock(&_sem_enter);

		if (ret)
		{
			errs++;
			printf("FAIL <---  (%d. Semaphore 'enter' lock failure.)\n", __LINE__);
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Wait for the thread to become "waiting".
		errs	= _wait_for_waiting(fd, ptr);

		if (errs)
		{
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Initiate the interrupt.
		errs			= ptr->func(fd, 0);

		if (errs)
		{
			printf("FAIL <---  (%d. Interrupt function failure.)\n", __LINE__);
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Wait for the thread to stop running.
		ret	= os_sem_lock(&_sem_exit);

		if (ret)
		{
			errs++;
			printf("FAIL <---  (%d. Semaphore 'exit' lock failure.)\n", __LINE__);
			_wait_cancel(fd);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// End processing.
		_wait_cancel(fd);	// just in case
		_wait_for_done();
		os_thread_destroy(&thread);

		if (_thread_err[0])
		{
			errs++;
			printf("FAIL <---  (%d. Failure in wait thread: %s)\n", __LINE__, _thread_err);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			errs++;
			printf("FAIL <---  (%d. Wait for interrupt timed out.)\n", __LINE__);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_CANCEL)
		{
			errs++;
			printf("FAIL <---  (%d. Wait for interrupt was cancelled.)\n", __LINE__);
			break;
		}

		if (_thread_wait.flags & ~GSC_WAIT_FLAG_DONE)
		{
			errs++;
			printf("FAIL <---  (%d. Invalid wait flags results: 0x%lX)\n", __LINE__, (long) _thread_wait.flags);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_DONE)
		{
			printf("PASS\n");
			break;
		}

		errs++;
		printf("FAIL <---  (%d. Invalid wait flags results: 0x%lX)\n", __LINE__, (long) _thread_wait.flags);
		break;
	}

	os_sem_destroy(&_sem_enter);
	os_sem_destroy(&_sem_exit);
	return(errs);
}



//*****************************************************************************
int perform_tests(const args_t* args)
{
	int	errs	= 0;
	int	i;

	for (i = 0; _tests[i].name; i++)
		errs	+= _test_irq(args->fd, &_tests[i]);

	return(errs);
}


