// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32-COS/HPDI32-COS-1.x_GSC_DN/utils/id.c $
// $Rev: 54389 $
// $Date: 2024-05-01 08:41:55 -0500 (Wed, 01 May 2024) $

// HPDI32-COS: Utilities: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_id_device_pci
*
*	Purpose:
*
*		Identify the device using the PCI registers.
*
*	Arguments:
*
*		fd		Use this handle to access the device.
*
*		verbose	Work verbosely? 0 = no, !0 = yes
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

static int _id_device_pci(int fd, int verbose)
{
	char	buf[128];
	u32		didr;
	int		errs	= 0;
	u32		reg;

	// Vendor ID ====================================================
	errs	+= hpdi32cos_reg_read(fd, -1, 0, GSC_PCI_9080_VIDR, &reg);
	sprintf(buf, "0x%04lX      ", (long) reg);

	if (reg == 0x10B5)
	{
		strcat(buf, "(PLX)\n");
	}
	else
	{
		errs	= 1;
		strcat(buf, "(UNKNOWN) FAIL <---\n");
	}

	if (verbose)
	{
		gsc_label("Vendor ID");
		printf("%s", buf);
	}

	// Device ID ====================================================
	errs	+= hpdi32cos_reg_read(fd, -1, 0, GSC_PCI_9080_DIDR, &didr);
	sprintf(buf, "0x%04lX      ", (long) didr);

	if (didr == 0x9080)
	{
		strcat(buf, "(PCI9080)\n");
	}
	else
	{
		errs	= 1;
		strcat(buf, "(UNKNOWN) FAIL <---\n");
	}

	if (verbose)
	{
		gsc_label("Device ID");
		printf("%s", buf);
	}

	// Subsystem Vendor ID ==========================================
	errs	+= hpdi32cos_reg_read(fd, -1, 0, GSC_PCI_9080_SVID, &reg);
	sprintf(buf, "0x%04lX      ", (long) reg);

	if (reg == 0x10B5)
	{
		strcat(buf, "(PLX)\n");
	}
	else
	{
		errs	= 1;
		strcat(buf, "(UNKNOWN) FAIL <---\n");
	}

	if (verbose)
	{
		gsc_label("Sub Vendor ID");
		printf("%s", buf);
	}

	// Subsystem ID =================================================
	errs	+= hpdi32cos_reg_read(fd, -1, 0, GSC_PCI_9080_SID, &reg);
	sprintf(buf, "0x%04lX      ", (long) reg);

	if (reg == 0x2400)
	{
		strcat(buf, "(HPDI32-COS)\n");
	}
	else
	{
		errs	= 1;
		strcat(buf, "(UNKNOWN) FAIL <---\n");
	}

	if (verbose)
	{
		gsc_label("Subsystem ID");
		printf("%s", buf);
	}

	return(errs);
}



//*****************************************************************************
static int _get_prefix(int fd, char* dst)
{
	int			errs;
	s32			ff;
	const char*	pre		= "";

	dst		+= strlen(dst);
	errs	= hpdi32cos_query(fd, -1, 0, HPDI32COS_QUERY_FORM_FACTOR, &ff);

	if (errs)
		sprintf(dst, " FAIL <---  (query error(s))");

	if (errs == 0)
	{
		switch (ff)
		{
			default:

				errs	= 1;
				sprintf(dst, " FAIL <---  (invalid form factor: %ld)", (long) ff);
				break;

			case HPDI32COS_FORM_FACTOR_UNKNOWN:					break;
			case HPDI32COS_FORM_FACTOR_PCI:		pre	= "PCI-";	break;
			case HPDI32COS_FORM_FACTOR_PMC:		pre	= "PMC-";	break;
		}
	}

	if (errs == 0)
		sprintf(dst, "%s", pre);

	return(errs);
}



//*****************************************************************************
static int _get_memory_size(int fd, char* dst)
{
	int			errs;
	const char*	factor		= "";
	s32			rx			= 0;

	dst		+= strlen(dst);
	errs	= hpdi32cos_query(fd, -1, 0, HPDI32COS_QUERY_FIFO_SIZE_RX, &rx);

	if (errs)
		sprintf(dst, " FAIL <---  (query error(s))");

	rx	*= 4;

	if ((rx) && ((rx % 1024) == 0))
	{
		rx		/= 1024;
		factor	= "K";
	}

	if ((rx) && ((rx % 1024) == 0))
	{
		rx		/= 1024;
		factor	= "M";
	}

	if ((errs == 0) && (rx > 0))
		sprintf(dst, "-%ld%s", (long) rx, factor);

	return(errs);
}



//*****************************************************************************
static int _get_model(int fd, int index, char* dst)
{
	int		errs		= 0;
	s32		model;
	s32		type;

	dst		+= strlen(dst);
	errs	+= hpdi32cos_query(fd, -1, 0, HPDI32COS_QUERY_DEVICE_TYPE, &type);
	errs	+= hpdi32cos_query(fd, -1, 0, HPDI32COS_QUERY_MODEL, &model);

	if (errs)
	{
		sprintf(dst, " FAIL <---  (query error(s))");
	}
	else if (type != GSC_DEV_TYPE_HPDI32COS)
	{
		errs	= 1;
		sprintf(dst, " FAIL <---  (invalid device type: %ld)", (long) type);
	}
	else
	{
		switch (model)
		{
			default:

				errs	= 1;
				strcat(dst, "UNKNOWN");
				break;

			case HPDI32COS_MODEL_HPDI32A:

				strcat(dst, "HPDI32A");
				break;

			case HPDI32COS_MODEL_HPDI32B:

				strcat(dst, "HPDI32B");
				break;
		}
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	hpdi32cos_id_device
*
*	Purpose:
*
*		Identify the device.
*
*	Arguments:
*
*		fd		The handle to use to access the device.
*
*		index	The index of the device or -1 if we don't care.
*
*		verbose	Work verbosely?
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int hpdi32cos_id_device(int fd, int index, int verbose)
{
	int		errs;
	char	model[0124]		= "";

	if (verbose)
		gsc_label_index("Device", index);

	errs	= _get_prefix(fd, model);

	if (errs == 0)
		errs	= _get_model(fd, index, model);

	if (errs == 0)
		errs	= _get_memory_size(fd, model);

	if (verbose)
	{
		printf("%s\n", model);
		gsc_label_level_inc();
	}

	errs	+= _id_device_pci(fd, verbose);

	if (verbose)
		gsc_label_level_dec();

	return(errs);
}



